#include <cmath>
#include <numeric>
#include <algorithm>
#include <stdexcept>

#include "npstat/nm/StorablePolySeries1D.hh"
#include "npstat/nm/GaussLegendreQuadrature.hh"

namespace npstat {
    namespace Private {
        class ClassicalPolySeries : public Functor1<double, double>
        {
        public:
            inline ClassicalPolySeries(const AbsClassicalOrthoPoly1D& p,
                                       const double *coeffs, const unsigned maxdeg)
                : poly_(p), coeffs_(coeffs), maxdeg_(maxdeg) {}

            inline double operator()(const double& x) const
            {
                return poly_.series(coeffs_, maxdeg_, x);
            }

        private:
            const AbsClassicalOrthoPoly1D& poly_;
            const double* coeffs_;
            unsigned maxdeg_;
        };

        class ClassicalPolyPower : public Functor1<long double, long double>
        {
        public:
            inline ClassicalPolyPower(const AbsClassicalOrthoPoly1D& p,
                                      const unsigned deg, const unsigned power,
                                      const bool i_useWeight)
                : poly(p), deg1(deg), polypow(power), useWeight(i_useWeight) {}

            inline long double operator()(const long double& x) const
            {
                long double p = 1.0L, w = 1.0L;
                if (useWeight)
                    w = poly.weight(x);
                if (polypow)
                {
                    p = poly.poly(deg1, x);
                    switch (polypow)
                    {
                    case 1U:
                        break;
                    case 2U:
                        p *= p;
                        break;
                    default:
                        p = powl(p, polypow);
                        break;
                    }
                }
                return w*p;
            }

        private:
            const AbsClassicalOrthoPoly1D& poly;
            unsigned deg1;
            unsigned polypow;
            bool useWeight;
        };
    }

    long double AbsClassicalOrthoPoly1D::poly(
        const unsigned degree, const long double x) const
    {
        long double polyk = 1.0L;
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::poly: degree argument is out of range");
            long double polykm1 = 0.0L;
            std::pair<long double,long double> rcurrent = recurrenceCoeffs(0);
            for (unsigned k=0; k<degree; ++k)
            {
                const std::pair<long double,long double>& rnext = recurrenceCoeffs(k+1);
                const long double p = ((x - rcurrent.first)*polyk -
                                       rcurrent.second*polykm1)/rnext.second;
                polykm1 = polyk;
                polyk = p;
                rcurrent = rnext;
            }
        }
        return polyk;
    }

    long double AbsClassicalOrthoPoly1D::monic(
        const unsigned degree, const long double x) const
    {
        long double polyk = 1.0L;
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::monic: degree argument is out of range");
            long double polykm1 = 0.0L;
            for (unsigned k=0; k<degree; ++k)
            {
                const std::pair<long double,long double>& rcurrent = monicRecurrenceCoeffs(k);
                const long double p = (x - rcurrent.first)*polyk - rcurrent.second*polykm1;
                polykm1 = polyk;
                polyk = p;
            }
        }
        return polyk;
    }

    long double AbsClassicalOrthoPoly1D::normpolyprod(
        const unsigned* degrees, const unsigned nDegrees,
        const unsigned maxdeg, const long double x) const
    {
        long double prod = 1.0L;
        if (nDegrees)
        {
            assert(degrees);
            if (maxdeg)
            {
                if (maxdeg > maxDegree()) throw std::invalid_argument(
                    "In npstat::AbsClassicalOrthoPoly1D::normpolyprod: "
                    "maximum degree argument is out of range");
                long double polyk = 1.0L, polykm1 = 0.0L;
                std::pair<long double,long double> rcurrent = recurrenceCoeffs(0);
                for (unsigned k=0; k<maxdeg; ++k)
                {
                    for (unsigned j=0; j<nDegrees; ++j)
                        if (k == degrees[j])
                            prod *= polyk;
                    const std::pair<long double,long double>& rnext = recurrenceCoeffs(k+1);
                    const long double p = ((x - rcurrent.first)*polyk -
                                           rcurrent.second*polykm1)/rnext.second;
                    polykm1 = polyk;
                    polyk = p;
                    rcurrent = rnext;
                }
                for (unsigned j=0; j<nDegrees; ++j)
                    if (maxdeg == degrees[j])
                        prod *= polyk;
            }
        }
        return prod;
    }

    std::pair<long double,long double>
    AbsClassicalOrthoPoly1D::twopoly(
        const unsigned deg1, const unsigned deg2, const long double x) const
    {
        long double p1 = 1.0L, p2 = 1.0L;
        const unsigned degree = std::max(deg1, deg2);
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::twopoly: "
                "degree argument is out of range");
            long double polyk = 1.0L, polykm1 = 0.0L;
            std::pair<long double,long double> rcurrent = recurrenceCoeffs(0);
            for (unsigned k=0; k<degree; ++k)
            {
                if (k == deg1)
                    p1 = polyk;
                if (k == deg2)
                    p2 = polyk;
                const std::pair<long double,long double>& rnext = recurrenceCoeffs(k+1);
                const long double p = ((x - rcurrent.first)*polyk -
                                       rcurrent.second*polykm1)/rnext.second;
                polykm1 = polyk;
                polyk = p;
                rcurrent = rnext;
            }
            if (deg1 == degree)
                p1 = polyk;
            if (deg2 == degree)
                p2 = polyk;
        }
        return std::pair<long double,long double>(p1, p2);
    }

    Matrix<long double>
    AbsClassicalOrthoPoly1D::jacobiMatrix(const unsigned n) const
    {
        if (!n) throw std::invalid_argument(
            "In npstat::AbsClassicalOrthoPoly1D::jacobiMatrix: "
            "can not build matrix of zero size");
        if (n > this->maxDegree()) throw std::invalid_argument(
            "In npstat::AbsClassicalOrthoPoly1D::jacobiMatrix: "
            "matrix size is out of range");
        Matrix<long double> mat(n, n, 0);
        unsigned ip1 = 1;
        std::pair<long double,long double> rip1 = this->recurrenceCoeffs(0);
        for (unsigned i=0; i<n; ++i, ++ip1)
        {
            std::pair<long double,long double> rCoeffs = rip1;
            mat[i][i] = rCoeffs.first;
            if (i)
                mat[i][i-1] = rCoeffs.second;
            if (ip1 < n)
            {
                rip1 = this->recurrenceCoeffs(ip1);
                mat[i][ip1] = rip1.second;
            }
        }
        return mat;
    }

    double AbsClassicalOrthoPoly1D::series(const double *coeffs,
                                           const unsigned degree,
                                           const double xIn) const
    {
        assert(coeffs);
        long double sum = coeffs[0];
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::series: "
                "degree argument is out of range");
            const long double x = xIn;
            long double polyk = 1.0L, polykm1 = 0.0L;
            std::pair<long double,long double> rcurrent = recurrenceCoeffs(0);
            for (unsigned k=0; k<degree; ++k)
            {
                const std::pair<long double,long double>& rnext = recurrenceCoeffs(k+1);
                const long double p = ((x - rcurrent.first)*polyk -
                                       rcurrent.second*polykm1)/rnext.second;
                sum += p*coeffs[k+1];
                polykm1 = polyk;
                polyk = p;
                rcurrent = rnext;
            }
        }
        return sum;
    }

    double AbsClassicalOrthoPoly1D::integrateSeries(const double *coeffs,
                                                    const unsigned degree) const
    {
        assert(coeffs);
        if (degree)
        {
            const unsigned nPt = GaussLegendreQuadrature::minimalExactRule(degree);
            if (!nPt) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::integrateSeries: "
                "the number of coefficients is too large");
            GaussLegendreQuadrature quad(nPt);
            Private::ClassicalPolySeries fcn(*this, coeffs, degree);
            return quad.integrate(fcn, xmin(), xmax());
        }
        else
            return coeffs[0]*(xmax() - xmin());
    }

    void AbsClassicalOrthoPoly1D::allpoly(const long double x,
                                          long double* values,
                                          const unsigned degree) const
    {
        assert(values);
        values[0] = 1.0L;
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::allpoly: "
                "degree argument is out of range");
            long double polyk = 1.0L, polykm1 = 0.0L;
            std::pair<long double,long double> rcurrent = recurrenceCoeffs(0);
            for (unsigned k=0; k<degree; ++k)
            {
                const std::pair<long double,long double>& rnext = recurrenceCoeffs(k+1);
                const long double p = ((x - rcurrent.first)*polyk -
                                       rcurrent.second*polykm1)/rnext.second;
                polykm1 = polyk;
                polyk = p;
                rcurrent = rnext;
                values[k+1] = p;
            }
        }
    }

    void AbsClassicalOrthoPoly1D::allmonic(const long double x,
                                           long double* values,
                                           const unsigned degree) const
    {
        assert(values);
        values[0] = 1.0L;
        if (degree)
        {
            if (degree > maxDegree()) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::allmonic: "
                "degree argument is out of range");
            long double polyk = 1.0L, polykm1 = 0.0L;
            for (unsigned k=0; k<degree; ++k)
            {
                const std::pair<long double,long double>& rcurrent = monicRecurrenceCoeffs(k);
                const long double p = (x - rcurrent.first)*polyk - rcurrent.second*polykm1;
                polykm1 = polyk;
                polyk = p;
                values[k+1] = p;
            }
        }
    }

    double AbsClassicalOrthoPoly1D::jointIntegral(
        const unsigned* degrees, const unsigned nDegrees) const
    {
        if (nDegrees)
        {
            assert(degrees);
            const unsigned degSum = std::accumulate(degrees, degrees+nDegrees, 0U);
            const unsigned nPt = GaussLegendreQuadrature::minimalExactRule(degSum);
            if (!nPt) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::jointIntegral: "
                "joint poly degree is too high");
            GaussLegendreQuadrature quad(nPt);
            MultiProdFcn fcn(*this, degrees, nDegrees, false);
            return quad.integrate(fcn, xmin(), xmax());
        }
        else
            return xmax() - xmin();
    }

    double AbsClassicalOrthoPoly1D::integratePoly(
        const unsigned degree, const unsigned power) const
    {
        const unsigned degSum = degree*power;
        if (degSum)
        {
            const unsigned nPt = GaussLegendreQuadrature::minimalExactRule(degSum);
            if (!nPt) throw std::invalid_argument(
                "In npstat::AbsClassicalOrthoPoly1D::integratePoly: "
                "poly degree or power is too high");
            GaussLegendreQuadrature quad(nPt);
            Private::ClassicalPolyPower fcn(*this, degree, power, false);
            return quad.integrate(fcn, xmin(), xmax());
        }
        else
            return xmax() - xmin();
    }

    Matrix<double>
    AbsClassicalOrthoPoly1D::empiricalKroneckerMatrix(
        const AbsIntervalQuadrature1D& quad, const unsigned maxdeg) const
    {
        if (maxdeg > maxDegree()) throw std::invalid_argument(
            "In npstat::AbsClassicalOrthoPoly1D::empiricalKroneckerMatrix: "
            "requested matrix size is out of range");
        const long double xminld = xmin();
        const long double xmaxld = xmax();
        const long double midpoint = (xminld + xmaxld)/2.0L;
        const long double unit = (xmaxld - xminld)/2.0L;
        const unsigned nIntegrationPoints = quad.npoints();
        std::vector<long double> ax(nIntegrationPoints);
        std::vector<long double> wx(nIntegrationPoints);
        quad.getAllAbscissae(&ax[0], ax.size());
        quad.getAllWeights(&wx[0], wx.size());
        const unsigned maxdegP1 = maxdeg + 1U;
        std::vector<long double> polyBuffer(maxdegP1);
        long double* const polyValues = &polyBuffer[0];
        Matrix<long double> acc(maxdegP1, maxdegP1, 0);
        for (unsigned ipt=0; ipt<nIntegrationPoints; ++ipt)
        {
            const long double x = midpoint + ax[ipt]*unit;
            const long double w = wx[ipt] * this->weight(x);
            allpoly(x, polyValues, maxdeg);
            for (unsigned irow=0; irow<=maxdeg; ++irow)
            {
                long double* row = acc[irow];
                const long double pvalrow = w*polyValues[irow];
                for (unsigned icol=0; icol<=maxdeg; ++icol)
                    row[icol] += pvalrow*polyValues[icol];
            }
        }
        acc *= unit;
        return acc;
    }

    CPP11_auto_ptr<StorablePolySeries1D> AbsClassicalOrthoPoly1D::makeStorablePolySeries(
        const unsigned maxPolyDeg, const double *coeffs,
        const unsigned maxdeg, const double shift) const
    {
        if (maxPolyDeg > maxDegree()) throw std::invalid_argument(
            "In npstat::AbsClassicalOrthoPoly1D::makeStorablePolySeries: "
            "poly degree argument is out of range");
        std::vector<std::pair<long double,long double> > rc(maxPolyDeg+1U);
        for (unsigned i=0; i<=maxPolyDeg; ++i)
            rc[i] = this->recurrenceCoeffs(i);
        return CPP11_auto_ptr<StorablePolySeries1D>(
            new StorablePolySeries1D(rc, this->xmin(), this->xmax(), shift, coeffs, maxdeg));        
    }
}
