#ifndef NPSTAT_CHEBYSHEVINTEGRAL_HH_
#define NPSTAT_CHEBYSHEVINTEGRAL_HH_

/*!
// \file ChebyshevIntegral.hh
//
// \brief Integrate functions by representing them with Chebyshev series
//
// Author: I. Volobouev
//
// March 2023
*/

#include <vector>
#include <cassert>

#include "npstat/nm/SimpleFunctors.hh"
#include "npstat/nm/MathUtils.hh"

namespace npstat {
    class ChebyshevIntegral : public Functor1<long double,long double>
    {
    public:
        template<class Functor>
        inline ChebyshevIntegral(const Functor& fcn, const unsigned maxdeg,
                                 const long double xmin, const long double xmax)
            : xmin_(xmin), xmax_(xmax), coeffs_(maxdeg+2U)
        {
            assert(xmin < xmax);
            std::vector<long double> series(maxdeg+1U);
            chebyshevSeriesCoeffs(fcn, xmin, xmax, maxdeg, &series[0]);
            chebyshevIntegralCoeffs(&series[0], maxdeg, xmin, xmax, &coeffs_[0]);
        }

        inline virtual ~ChebyshevIntegral() {}

        // x must lie inside the interval [xmin, xmax],
        // where xmin and xmax were specified in the constructor.
        // This operator returns the value of the integral
        // from xmin to x.
        inline long double operator()(const long double& x) const
        {
            assert(x >= xmin_ && x <= xmax_);
            return chebyshevSeriesSum(&coeffs_[0], coeffs_.size()-1U,
                                      xmin_, xmax_, x);
        }

    private:
        long double xmin_;
        long double xmax_;
        std::vector<long double> coeffs_;
    };
}

#endif // NPSTAT_CHEBYSHEVINTEGRAL_HH_
