#include <cmath>
#include <cassert>
#include <stdexcept>
#include <algorithm>

#include "npstat/nm/SpecialFunctions.hh"
#include "npstat/nm/GaussLegendreQuadrature.hh"
#include "npstat/nm/ClassicalOrthoPolys1D.hh"

#include "npstat/rng/permutation.hh"

// Calculate (-1)^k
static inline int powm1(const unsigned k)
{
    return k % 2U ? -1 : 1;
}

namespace {
    class PolyProdWithSquaredPoly : public npstat::Functor1<long double, long double>
    {
    public:
        inline PolyProdWithSquaredPoly(const npstat::AbsClassicalOrthoPoly1D& fcn,
                                       const unsigned deg1, const unsigned deg2)
            : fcn_(fcn), degree1_(deg1), degree2_(deg2) {}

        inline virtual ~PolyProdWithSquaredPoly() {}

        inline virtual long double operator()(const long double& a) const
        {
            std::pair<long double,long double> p = fcn_.twopoly(
                degree1_, degree2_, a);
            return p.first * p.second * p.second;
        }

    private:
        const npstat::AbsClassicalOrthoPoly1D& fcn_;
        unsigned degree1_;
        unsigned degree2_;
    };
}

namespace npstat {
    void LegendreOrthoPoly1D::monomialCoeffs(
        const double* coeffs, const unsigned degree,
        long double* monoCoeffs) const
    {
        assert(coeffs);
        assert(monoCoeffs);

        std::fill(monoCoeffs, monoCoeffs+(degree+1U), 0.0L);

        for (unsigned deg=0U; deg<=degree; ++deg)
            if (coeffs[deg])
            {
                const long double norm = coeffs[deg]*sqrtl(2.0L*deg + 1)/powl(2.0L, deg);
                const unsigned kmax = deg/2U;
                for (unsigned k=0; k<=kmax; ++k)
                    monoCoeffs[deg-2U*k] += norm*powm1(k)*ldfactorial(2U*(deg-k))/ldfactorial(k)/ldfactorial(deg-k)/ldfactorial(deg-2U*k);
            }
    }

    void LegendreOrthoPoly1D::integralCoeffs(
        const double* coeffs, const unsigned degree,
        double* integCoeffs) const
    {
        const double sqr3 = 1.7320508075688773;

        assert(coeffs);
        assert(integCoeffs);

        integCoeffs[0] = coeffs[0];
        integCoeffs[1] = coeffs[0]/sqr3;
        std::fill(integCoeffs+2U, integCoeffs+(degree+2U), 0.0);

        for (unsigned deg=1U; deg<=degree; ++deg)
            if (coeffs[deg])
            {
                const double tmp = coeffs[deg]/sqrt(2.0*deg + 1);
                integCoeffs[deg+1U] += tmp/sqrt(2.0*deg + 3);
                integCoeffs[deg-1U] -= tmp/sqrt(2.0*deg - 1);
            }
    }

    void LegendreOrthoPoly1D::derivativeCoeffs(
        const double* coeffs, const unsigned degree,
        double* derivCoeffs) const
    {
        if (degree)
        {
            assert(coeffs);
            assert(derivCoeffs);

            std::fill(derivCoeffs, derivCoeffs+degree, 0.0);

            for (unsigned deg=1U; deg<=degree; ++deg)
                if (coeffs[deg])
                {
                    const double tmp = coeffs[deg]*sqrt(2.0*deg + 1);
                    for (unsigned k=0U; k<deg; k+=2U)
                        derivCoeffs[deg-1U-k] += tmp*sqrt(2.0*(deg-k) - 1);
                }
        }
        else
        {
            if (derivCoeffs)
                derivCoeffs[0] = 0.0;
        }
    }

    void LegendreOrthoPoly1D::squaredPolyCoeffs(
        const unsigned degree, long double* expansionCoeffs) const
    {
        assert(expansionCoeffs);
        expansionCoeffs[0] = 2.0;

        if (degree)
        {
            const unsigned nIntegPt = GaussLegendreQuadrature::minimalExactRule(4U*degree);
            assert(nIntegPt);
            const GaussLegendreQuadrature glq(nIntegPt);

            for (unsigned deg=1U; deg<=2U*degree; ++deg)
            {
                if (deg % 2U)
                    expansionCoeffs[deg] = 0.0;
                else
                {
                    PolyProdWithSquaredPoly prod(*this, deg, degree);
                    expansionCoeffs[deg] = glq.integrate(prod, -1.0, 1.0);
                }
            }
        }
    }

    void ShiftedLegendreOrthoPoly1D::integralCoeffs(
        const double* coeffs, const unsigned degree,
        double* integCoeffs) const
    {
        const double twosqr3 = 3.4641016151377546;

        assert(coeffs);
        assert(integCoeffs);

        integCoeffs[0] = coeffs[0]/2.0;
        integCoeffs[1] = coeffs[0]/twosqr3;
        std::fill(integCoeffs+2U, integCoeffs+(degree+2U), 0.0);

        for (unsigned deg=1U; deg<=degree; ++deg)
            if (coeffs[deg])
            {
                const double tmp = coeffs[deg]/2.0/sqrt(2.0*deg + 1);
                integCoeffs[deg+1U] += tmp/sqrt(2.0*deg + 3);
                integCoeffs[deg-1U] -= tmp/sqrt(2.0*deg - 1);
            }
    }

    void ShiftedLegendreOrthoPoly1D::derivativeCoeffs(
        const double* coeffs, const unsigned degree,
        double* derivCoeffs) const
    {
        if (degree)
        {
            assert(coeffs);
            assert(derivCoeffs);

            std::fill(derivCoeffs, derivCoeffs+degree, 0.0);

            for (unsigned deg=1U; deg<=degree; ++deg)
                if (coeffs[deg])
                {
                    const double tmp = 2.0*coeffs[deg]*sqrt(2.0*deg + 1);
                    for (unsigned k=0U; k<deg; k+=2U)
                        derivCoeffs[deg-1U-k] += tmp*sqrt(2.0*(deg-k) - 1);
                }
        }
        else
        {
            if (derivCoeffs)
                derivCoeffs[0] = 0.0;
        }
    }

    void ShiftedLegendreOrthoPoly1D::squaredPolyCoeffs(
        const unsigned degree, long double* expansionCoeffs) const
    {
        assert(expansionCoeffs);
        expansionCoeffs[0] = 1.0;

        if (degree)
        {
            const unsigned nIntegPt = GaussLegendreQuadrature::minimalExactRule(4U*degree);
            assert(nIntegPt);
            const GaussLegendreQuadrature glq(nIntegPt);

            for (unsigned deg=1U; deg<=2U*degree; ++deg)
            {
                if (deg % 2U)
                    expansionCoeffs[deg] = 0.0;
                else
                {
                    PolyProdWithSquaredPoly prod(*this, deg, degree);
                    expansionCoeffs[deg] = glq.integrate(prod, 0.0, 1.0);
                }
            }
        }
    }

    JacobiOrthoPoly1D::JacobiOrthoPoly1D(const double a, const double b)
        : alpha_(a),
          beta_(b)
    {
        if (!(alpha_ > -1.0 && beta_ > -1.0)) throw std::invalid_argument(
            "In npstat::JacobiOrthoPoly1D constructor: invalid arguments");
        norm_ = powl(2.0L, alpha_+beta_+1)*Gamma(alpha_+1)*Gamma(beta_+1)/
            Gamma(alpha_+beta_+2);
    }

    long double JacobiOrthoPoly1D::weight(const long double x) const
    {
        if (x < -1.0L || x > 1.0L)
            return 0.0L;
        else
        {
            if (x == 1.0L && alpha_ < 0.0L) throw std::invalid_argument(
                "In npstat::JacobiOrthoPoly1D::weight: invalid x=1 argument");
            if (x == -1.0L && beta_ < 0.0L) throw std::invalid_argument(
                "In npstat::JacobiOrthoPoly1D::weight: invalid x=-1 argument");

            // Don't call "powl" if we can avoid it because this is a slow function
            long double p;
            if (alpha_ == beta_)
            {
                const long double tmp = 1.0L - x*x;

                // Fourth power is a special case, often used in density estimation
                if (alpha_ == 4.0)
                {
                    const long double tmpsq = tmp*tmp;
                    p = tmpsq*tmpsq;
                }
                else if (alpha_ == 0.0)
                    p = 1.0L;
                else if (alpha_ == 1.0)
                    p = tmp;
                else if (alpha_ == 2.0)
                    p = tmp*tmp;
                else if (alpha_ == 3.0)
                    p = tmp*tmp*tmp;
                else if (alpha_ == 5.0)
                {
                    const long double tmpsq = tmp*tmp;
                    p = tmpsq*tmpsq*tmp;
                }
                else if (alpha_ == 6.0)
                {
                    const long double tmpsq = tmp*tmp;
                    p = tmpsq*tmpsq*tmpsq;
                }
                else
                    p = powl(tmp, alpha_);
            }
            else
                p = powl(1.0L - x, alpha_)*powl(1.0L + x, beta_);
            return p/norm_;
        }
    }

    std::pair<long double,long double>
    JacobiOrthoPoly1D::monicRecurrenceCoeffs(const unsigned k) const
    {
        long double a, b = 1.0L;
        if (k)
        {
            const long double tmp = 2*k + alpha_ + beta_;
            a = (beta_ - alpha_)*(beta_ + alpha_)/tmp/(tmp + 2);

            if (k == 1U)
                b = 4*k*(k+alpha_)*(k+beta_)/tmp/tmp/(tmp + 1);
            else
                b = 4*k*(k+alpha_)*(k+beta_)*(k+alpha_+beta_)/
                    tmp/tmp/(tmp + 1)/(tmp - 1);
        }
        else
            a = (beta_ - alpha_)/(beta_ + alpha_ + 2);
        return std::pair<long double,long double>(a, b);
    }

    std::pair<long double,long double>
    JacobiOrthoPoly1D::recurrenceCoeffs(const unsigned k) const
    {
        const std::pair<long double,long double>& p = this->monicRecurrenceCoeffs(k);
        return std::pair<long double,long double>(p.first, sqrtl(p.second));
    }

    void JacobiOrthoPoly1D::derivativeCoeffs(
        const double* coeffs, const unsigned maxdeg, double* derivCoeffs) const
    {
        if (maxdeg)
        {
            assert(coeffs);
            assert(derivCoeffs);
            assert(derivCoeffs != coeffs);

            const double aplusbplusone = alpha_ + beta_+ 1.0;
            const double commonf = 0.25/(alpha_+1.0)/(beta_+1.0)*(aplusbplusone+1.0)*(aplusbplusone+2.0);
            for (unsigned ideg=maxdeg; ideg>0U; --ideg)
                derivCoeffs[ideg-1U] = coeffs[ideg]*sqrt(ideg*(aplusbplusone + ideg)*commonf);
        }
    }

    long double HermiteProbOrthoPoly1D::weight(const long double x) const
    {
        const long double norm = 0.39894228040143267793994606L; // 1/sqrt(2 Pi)
        return norm*expl(-0.5*x*x);
    }

    std::pair<long double,long double>
    HermiteProbOrthoPoly1D::recurrenceCoeffs(const unsigned k) const
    {
        return std::pair<long double,long double>(0.0L, sqrtl(k));
    }

    std::pair<long double,long double>
    HermiteProbOrthoPoly1D::monicRecurrenceCoeffs(const unsigned k) const
    {
        return std::pair<long double,long double>(0.0L, k);
    }
}
