#include <cassert>
#include <algorithm>

#include "npstat/nm/ClassicalOrthoPolys1D.hh"
#include "npstat/nm/SimpleScalarProduct.hh"
#include "npstat/nm/LegendreCDTruncation.hh"

namespace npstat {
    LegendreCDTruncation::LegendreCDTruncation()
    {
        const std::vector<long double> v0(1, 1.0L);
        squaredPolyCoeffs_.push_back(v0);
    }

    void LegendreCDTruncation::fillCoeffs(const unsigned lambda) const
    {
        const unsigned currentLambda = squaredPolyCoeffs_.size() - 1U;
        if (lambda > currentLambda)
        {
            ShiftedLegendreOrthoPoly1D poly;
            for (unsigned deg = currentLambda+1U; deg <= lambda; ++deg)
            {
                std::vector<long double> buf(2U*deg + 1U);
                poly.squaredPolyCoeffs(deg, &buf[0]);
                squaredPolyCoeffs_.push_back(buf);
            }
        }
    }

    void LegendreCDTruncation::calcSeriesCoeffs(
        double* buf, const unsigned lambda) const
    {
        fillCoeffs(lambda);
        const unsigned buflen = 2U*lambda + 1U;
        for (unsigned ideg=0; ideg<buflen; ++ideg)
        {
            if (ideg % 2U)
                buf[ideg] = 0.0;
            else
            {
                long double sum = 0.0L;
                for (unsigned j=lambda; j>0; --j)
                {
                    if (ideg < squaredPolyCoeffs_[j].size())
                        sum += squaredPolyCoeffs_[j][ideg];
                    else
                        break;
                }
                buf[ideg] = sum;
            }
        }
    }

    double LegendreCDTruncation::calculateR(
        const double* coeffs, const unsigned lambda,
        const unsigned n) const
    {
        const unsigned buflen = 2U*lambda + 1U;
        if (dbuf_.size() < buflen)
            dbuf_.resize(buflen);
        calcSeriesCoeffs(&dbuf_[0], lambda);
        SimpleScalarProduct<double> sp;
        return 2.0*(sp(coeffs, &dbuf_[1], buflen-1U) + dbuf_[0])
            - (n+1U)*sp(coeffs, coeffs, lambda);
    }

    unsigned LegendreCDTruncation::bestLambda(
        const double* coeffs, const unsigned maxLambda,
        const unsigned n) const
    {
        assert(maxLambda);
        if (maxLambda == 1U)
            return 1U;
        else
        {
            std::vector<double> rbuf(maxLambda);
            for (unsigned lam=1; lam<=maxLambda; ++lam)
                rbuf[lam-1U] = calculateR(coeffs, lam, n);
            return std::distance(rbuf.begin(), std::min_element(rbuf.begin(), rbuf.end())) + 1U;
        }
    }
}
