#ifndef NPSTAT_LEGENDRECDTRUNCATION_HH_
#define NPSTAT_LEGENDRECDTRUNCATION_HH_

/*!
// \file LegendreCDTruncation.hh
//
// \brief Optimal truncation degree for OSDE by Legendre polynomials
//        on [0, 1] (typical for OSDE of a comparison density)
//
// Author: I. Volobouev
//
// March 2023
*/

#include <vector>

namespace npstat {
    class LegendreCDTruncation
    {
    public:
        LegendreCDTruncation();

        // The length of array "coeffs" should be at least 2*lambda.
        // The coefficients start with degree 1 (it is assumed that
        // the coefficient for degree 0 is 1). n is the sample size.
        double calculateR(const double* coeffs, unsigned lambda,
                          unsigned n) const;

        // The length of array "coeffs" should be at least 2*maxLambda.
        // The coefficients start with degree 1 (it is assumed that
        // the coefficient for degree 0 is 1). The best lambda will
        // minimize R calculated by "calculateR". Basically, this
        // returns optimal OSDE truncation degree according to the
        // Hart scheme. n is the sample size.
        unsigned bestLambda(const double* coeffs, unsigned maxLambda,
                            unsigned n) const;

    private:
        void fillCoeffs(unsigned lambda) const;

        // The length of "buf" array should be at least 2*lambda + 1
        void calcSeriesCoeffs(double* buf, unsigned lambda) const;

        mutable std::vector<std::vector<long double> > squaredPolyCoeffs_;
        mutable std::vector<double> dbuf_;
    };
}

#endif // NPSTAT_LEGENDRECDTRUNCATION_HH_
