#ifndef NPSTAT_KERNELSENSITIVITYMATRIX_HH_
#define NPSTAT_KERNELSENSITIVITYMATRIX_HH_

/*!
// \file kernelSensitivityMatrix.hh
//
// \brief Calculate kernel sensitivity matrix for KDE-like density estimation
//
// Author: I. Volobouev
//
// November 2022
*/

#include "npstat/nm/Matrix.hh"
#include "npstat/nm/AbsIntervalQuadrature1D.hh"
#include "npstat/nm/ContOrthoPoly1D.hh"

namespace npstat {
    /**
    // Template class Fcn2D should provide a method
    // "Real operator()(Real x, Real y) const", where "Real" is one of
    // floating point types (long double works best).
    //
    // It is assumed that the interfaces of classes "InPoly" and "OutPoly"
    // are similar to those of npstat classes AbsClassicalOrthoPoly1D or
    // ScalableClassicalOrthoPoly1D.
    //
    // In the returned matrix, row numbers correspond to the "output"
    // polynomial degrees (y space) and column numbers to the "input"
    // polynomial degrees (x space).
    //
    // Set "normalizeKernel" parameter to "false" if the kernel is already known
    // to be normalized, that is, Int_ymin^ymax K(x, y) dy = 1 for every x.
    */
    template <class Fcn2D, class InPoly, class OutPoly>
    Matrix<double> kernelSensitivityMatrix(
        const Fcn2D& kernel,
        const InPoly& inputPoly, unsigned maxdegInputPoly,
        const OutPoly& outPoly, unsigned maxdegOutPoly,
        const AbsIntervalQuadrature1D& xIntegrator,
        const AbsIntervalQuadrature1D& yIntegrator,
        bool normalizeKernel=true);

    /**
    // Template class Fcn2D should provide a method
    // "Real operator()(Real x, Real y) const", where "Real" is one of
    // floating point types (long double works best).
    //
    // It is assumed that the interface of class "OutPoly" is similar to that of
    // npstat classes AbsClassicalOrthoPoly1D or ScalableClassicalOrthoPoly1D.
    //
    // In the returned matrix, row numbers correspond to the "output"
    // polynomial degrees (y space) and column numbers to the "input"
    // polynomial degrees (x space).
    //
    // Set "normalizeKernel" parameter to "false" if the kernel is already known
    // to be normalized, that is, Int_ymin^ymax K(x, y) dy = 1 for every x.
    */
    template <class Fcn2D, class OutPoly>
    Matrix<double> kernelSensitivityMatrix(
        const Fcn2D& kernel,
        const ContOrthoPoly1D& inputPoly, unsigned maxdegInputPoly,
        const OutPoly& outPoly, unsigned maxdegOutPoly,
        const AbsIntervalQuadrature1D& yIntegrator,
        bool normalizeKernel=true);
}

#include "npstat/nm/kernelSensitivityMatrix.icc"

#endif // NPSTAT_KERNELSENSITIVITYMATRIX_HH_
