#ifndef NPSTAT_MODIFIEDGRAMSCHMIDT_HH_
#define NPSTAT_MODIFIEDGRAMSCHMIDT_HH_

/*!
// \file modifiedGramSchmidt.hh
//
// \brief Numerically stable modified Gram-Schmidt procedure
//
// Author: I. Volobouev
//
// February 2023
*/

namespace npstat {
    /**
    // Numerically stable modified Gram-Schmidt procedure.
    // Builds the orthogonal vectors in-place. All rows
    // will be orthonormal on exit.
    //
    // The input data should represent a collection of input
    // vectors, row-by-row. If this is treated as a matrix,
    // input parameter "nVectors" is the number of rows, and
    // "dim" is the number of columns. Normally, we must have
    // nVectors <= dim.
    //
    // SprodFunctor is a functor which should have a method
    // with the signature that looks something like
    // Real operator()(const Real*, const Real*, unsigned long) const
    //
    // The parameter "vector0AlreadyNormalized" should normally
    // be set to "true" if we are creating an orthogonal polynomial
    // system. For OPSs, the first vector should consist of 1s, while
    // the normalization is performed instead on the weight function.
    */
    template<typename Real, class SprodFunctor>
    void modifiedGramSchmidt(Real* data, unsigned nVectors, unsigned dim,
                             const SprodFunctor& sprodCalc,
                             bool vector0AlreadyNormalized=false);
}

#include "npstat/nm/modifiedGramSchmidt.icc"

#endif // NPSTAT_MODIFIEDGRAMSCHMIDT_HH_
