#include <cassert>
#include <cmath>
#include <vector>

#include "npstat/nm/scalesFromHessian.hh"

namespace npstat {
    void scalesFromHessian(const Matrix<double>& hessian,
                           const double negativeScaleLimit,
                           double* scales, const unsigned nScales)
    {
        assert(hessian.isSquare());
        assert(negativeScaleLimit >= 0.0);
        const unsigned nCoeffs = hessian.nRows();
        assert(nScales >= nCoeffs);
        
        for (unsigned i=0; i<nCoeffs; ++i)
        {
            const double h = hessian[i][i];
            if (h > 0.0)
                scales[i] = 1.0/std::sqrt(h);
            else if (h == 0.0)
                scales[i] = negativeScaleLimit;
            else
            {
                scales[i] = 1.0/std::sqrt(-h);
                if (scales[i] > negativeScaleLimit)
                    scales[i] = negativeScaleLimit;
            }
        }
    }

    Matrix<double> covmatFromHessian(const Matrix<double>& hessian,
                                     const double negativeScaleLimit)
    {
        assert(hessian.isSquare());
        const unsigned nCoeffs = hessian.nRows();
        const double squaredLimit = negativeScaleLimit*negativeScaleLimit;

        std::vector<double> hessianEigenvalues(nCoeffs);
        Matrix<double> hessianEigenvectors(nCoeffs, nCoeffs);
        hessian.symEigen(&hessianEigenvalues[0], nCoeffs, &hessianEigenvectors);
        for (unsigned i=0; i<nCoeffs; ++i)
        {
            if (hessianEigenvalues[i] > 0.0)
                hessianEigenvalues[i] = 1.0/hessianEigenvalues[i];
            else if (hessianEigenvalues[i] == 0.0)
                hessianEigenvalues[i] = squaredLimit;
            else
            {
                hessianEigenvalues[i] = -1.0/hessianEigenvalues[i];
                if (hessianEigenvalues[i] > squaredLimit)
                    hessianEigenvalues[i] = squaredLimit;
            }
        }
        return diag(&hessianEigenvalues[0], nCoeffs).bilinear(hessianEigenvectors);
    }
}
