#include <cmath>
#include <limits>
#include <cassert>
#include <stdexcept>
#include <algorithm>
#include <map>
#include <utility>

#include "npstat/rng/permutation.hh"
#include "npstat/nm/allocators.hh"
#include "npstat/nm/shiftedLegendreIntegral.hh"

namespace {
    class Ushort4key
    {
    public:
        // This assumes that "degs" array is sorted
        inline Ushort4key(const unsigned short degs[4])
        {
            degrees_[0] = degs[0];
            degrees_[1] = degs[1];
            degrees_[2] = degs[2];
            degrees_[3] = degs[3];
        }

        inline bool operator<(const Ushort4key& r) const
        {
            for (unsigned i=0; i<4; ++i)
            {
                if (degrees_[i] < r.degrees_[i])
                    return true;
                if (degrees_[i] > r.degrees_[i])
                    return false;
            }
            return false;
        }

    private:
        unsigned short degrees_[4];
    };
}

static long double threeJSquared(const unsigned i,
                                 const unsigned j,
                                 const unsigned k)
{
    const unsigned twos = i + j + k;
    if (twos % 2U) {return 0.0L;}
    if (2U*i > twos || 2U*j > twos || 2U*k > twos) {return 0.0L;}
    const unsigned s = twos/2U;
    const long double tmp1 = npstat::ldfactorial(s)/
                             npstat::ldfactorial(s - i)/
                             npstat::ldfactorial(s - j)/
                             npstat::ldfactorial(s - k);
    const long double tmp2 = npstat::ldfactorial(twos - 2U*i)*
                             npstat::ldfactorial(twos - 2U*j)*
                             npstat::ldfactorial(twos - 2U*k)/
                             npstat::ldfactorial(twos + 1U);
    return tmp2*tmp1*tmp1;
}

static void legendreProductCoeffs(const unsigned i, const unsigned j,
                                  long double* c, const unsigned toDeg)
{
    assert(c);
    const unsigned minDeg = i > j ? i - j : j - i;
    npstat::clearBuffer(c, std::min(minDeg, toDeg + 1U));

    const unsigned long tmp1 = static_cast<unsigned long>(2U*i + 1U)*(2U*j + 1U);
    for (unsigned m=minDeg; m<=toDeg; ++m)
    {
        const long double threeJSq = threeJSquared(i, j, m);
        if (threeJSq)
            c[m] = threeJSq*sqrtl(tmp1*(2U*m + 1U));
        else
            c[m] = 0.0L;
    }
}

namespace npstat {
    long double shiftedLegendreIntegral(const unsigned i,
                                        const unsigned j,
                                        const unsigned k)
    {
        if (!i) {return j == k ? 1.0L : 0.0L;}
        if (!j) {return i == k ? 1.0L : 0.0L;}
        if (!k) {return i == j ? 1.0L : 0.0L;}
        const long double threeJSq = threeJSquared(i, j, k);
        if (threeJSq)
        {
            const unsigned long tmp3 = static_cast<unsigned long>(2U*i + 1U)*
                                       (2U*j + 1U)*(2U*k + 1U);
            return threeJSq*sqrtl(tmp3);
        }
        else
            return 0.0L;
    }

    long double shiftedLegendreIntegral(const unsigned i, const unsigned j,
                                        const unsigned k, const unsigned l,
                                        const bool memoize)
    {
        typedef std::map<Ushort4key,long double> ResultMap;
        static ResultMap memoized;

        static const unsigned maxshort = std::numeric_limits<unsigned short>::max();
        if (i > maxshort || j > maxshort || k > maxshort || l > maxshort)
            throw std::invalid_argument(
                "In npstat::shiftedLegendreIntegral: polynomial degree is way too large");

        // The maximum argument of ldfactorial is 1754.
        // We need to have a buffer which is comparable.
        static const unsigned maxdeg = 4000U;

        const unsigned deg = i + j + k + l;
        if (deg % 2U) {return 0.0L;}
        if (2U*i > deg || 2U*j > deg || 2U*k > deg || 2U*l > deg) {return 0.0L;}
        if (!i) {return shiftedLegendreIntegral(j, k, l);}
        if (!j) {return shiftedLegendreIntegral(i, k, l);}
        if (!k) {return shiftedLegendreIntegral(i, j, l);}
        if (!l) {return shiftedLegendreIntegral(i, j, k);}

        // Among all possible pairs that can be made out of the
        // set {i, j, k, l}, choose the pair with the smallest
        // degree sum
        unsigned short degrees[4];
        degrees[0] = i;
        degrees[1] = j;
        degrees[2] = k;
        degrees[3] = l;
        std::sort(degrees, degrees + 4);
        const unsigned sumdeg = static_cast<unsigned>(degrees[0]) + degrees[1];

        if (2U*sumdeg > maxdeg) throw std::invalid_argument(
            "In npstat::shiftedLegendreIntegral: polynomial degree is too large");

        if (memoize)
        {
            ResultMap::const_iterator it = memoized.find(Ushort4key(degrees));
            if (it != memoized.end())
                return it->second;
        }

        long double buffer[maxdeg + 2U];
        long double* c1 = buffer;
        long double* c2 = c1 + (sumdeg + 1U);
        legendreProductCoeffs(degrees[0], degrees[1], c1, sumdeg);
        legendreProductCoeffs(degrees[2], degrees[3], c2, sumdeg);
        long double sum = 0.0L;
        for (unsigned q=0; q<=sumdeg; ++q)
            sum += c1[q]*c2[q];

        if (memoize)
            memoized.insert(std::make_pair(Ushort4key(degrees), sum));

        return sum;
    }
}
