#include <stdexcept>

#include "npstat/stat/CdfTransform1D.hh"

#include "geners/GenericIO.hh"
#include "geners/IOException.hh"

namespace npstat {
    void CdfTransform1D::validateScale(const double a)
    {
        if (a == 0.0) throw std::invalid_argument(
            "In npstat::CdfTransform1D::validateScale: "
            "scale parameter must not be zero");
    }

    CdfTransform1D::CdfTransform1D(const AbsDistribution1D& distro,
                                   const double a, const double b)
        : AbsDistributionTransform1D(2),
          distro_(distro.clone()), a_(a), b_(b)
    {
        validateScale(a_);
        xmin_ = distro_->quantile(0.0);
        xmax_ = distro_->quantile(1.0);
    }

    bool CdfTransform1D::write(std::ostream& os) const
    {
        gs::write_pod(os, a_);
        gs::write_pod(os, b_);
        return !os.fail() && gs::write_item(os, *distro_);
    }

    CdfTransform1D* CdfTransform1D::read(const gs::ClassId& id,
                                         std::istream& is)
    {
        static const gs::ClassId current(
            gs::ClassId::makeId<CdfTransform1D>());
        current.ensureSameId(id);

        double a, b;
        gs::read_pod(is, &a);
        gs::read_pod(is, &b);
        CPP11_auto_ptr<AbsDistribution1D> pd = gs::read_item<AbsDistribution1D>(is);
        if (is.fail()) throw gs::IOReadFailure(
            "In npstat::CdfTransform1D::read: input stream failure");
        return new CdfTransform1D(*pd, a, b);
    }

    double CdfTransform1D::transformForward(const double x, double* dydx) const
    {
        if (x < xmin_)
        {
            if (dydx) *dydx = 0.0;
            return b_;
        }
        else if (x > xmax_)
        {
            if (dydx) *dydx = 0.0;
            return a_ + b_;
        }
        else
        {
            if (dydx) *dydx = a_*distro_->density(x);
            return a_*distro_->cdf(x) + b_;
        }
    }

    double CdfTransform1D::transformBack(const double y) const
    {
        const double ymax = a_ + b_;
        if (y < b_ || y > ymax) throw std::domain_error(
            "In npstat::CdfTransform1D::transformBack: argument out of range");
        if (y == b_)
            return xmin_;
        else if (y == ymax)
            return xmax_;
        else
            return distro_->quantile((y - b_)/a_);
    }

    CdfTransform1D& CdfTransform1D::operator=(const CdfTransform1D& r)
    {
        if (&r != this)
        {
            AbsDistribution1D* tmp = r.distro_->clone();
            AbsDistributionTransform1D::operator=(r);
            delete distro_;
            distro_ = tmp;
            a_ = r.a_;
            b_ = r.b_;
            xmin_ = r.xmin_;
            xmax_ = r.xmax_;
        }
        return *this;
    }
}
