#include <cassert>
#include <stdexcept>
#include <sstream>
#include <climits>

#include "npstat/nm/findRootUsingBisections.hh"
#include "npstat/nm/Poly1D.hh"
#include "npstat/nm/MathUtils.hh"

#include "npstat/stat/ChebyshevDistro1D.hh"
#include "npstat/stat/distributionReadError.hh"

namespace npstat {
    ChebyshevDistro1D::ChebyshevDistro1D(
        const double location, const double scale,
        const double* coeffs, const unsigned maxdeg,
        const bool validatePositivity)
        : AbsScalableDistribution1D(location, scale)
    {
        setupCoeffs(coeffs, maxdeg);
        checkPositive(validatePositivity);
    }

    ChebyshevDistro1D::ChebyshevDistro1D(
        const double location, const double scale,
        const std::vector<double>& coeffs,
        const bool validatePositivity)
        : AbsScalableDistribution1D(location, scale)
    {
        setupCoeffs(coeffs.empty() ? (const double*)0 : &coeffs[0],
                    coeffs.size());
        checkPositive(validatePositivity);
    }

    void ChebyshevDistro1D::setupCoeffs(const double* coeffs,
                                        const unsigned maxdeg)
    {
        if (maxdeg)
            assert(coeffs);
        allCoeffs_.resize(maxdeg + 1U);
        long double integ = 0.0L;
        for (unsigned i=0; i<maxdeg; ++i)
        {
            const unsigned deg = i + 1U;
            allCoeffs_[deg] = coeffs[i];
            if (deg % 2U == 0U)
                integ += allCoeffs_[deg]*2.0L/(1.0L - static_cast<long double>(deg)*deg);
        }
        allCoeffs_[0] = static_cast<double>((1.0L - integ)/2.0L);
        for (unsigned i=maxdeg; i; --i)
        {
            if (allCoeffs_[i] == 0.0)
                allCoeffs_.pop_back();
            else
                break;
        }

        const unsigned deg = allCoeffs_.size() - 1;
        integCoeffs_.resize(deg + 2U);
        chebyshevIntegralCoeffs(&allCoeffs_[0], deg, -1.0, 1.0, &integCoeffs_[0]);
    }

    void ChebyshevDistro1D::checkPositive(const bool validatePositivity)
    {
        const unsigned sz = allCoeffs_.size();
        std::vector<long double> monoCoeffs(sz);
        chebyshevMonomialCoeffs(&allCoeffs_[0], sz-1U, &monoCoeffs[0]);
        Poly1D monoPoly(monoCoeffs);
        if (monoPoly.nRoots(-1.0L, 1.0L))
        {
            isPositive_ = false;
            if (validatePositivity) throw std::invalid_argument(
                "In npstat::ChebyshevDistro1D::checkPositive: "
                "the density is not strictly positive on [-1, 1]");
        }
        else
            isPositive_ = true;
    }

    double ChebyshevDistro1D::unscaledDensity(const double x) const
    {
        if (x < -1.0 || x > 1.0)
            return 0.0;
        else
            return chebyshevSeriesSum(&allCoeffs_[0], allCoeffs_.size()-1, x);
    }

    double ChebyshevDistro1D::unscaledCdf(const double x) const
    {
        if (x <= -1.0)
            return 0.0;
        else if (x >= 1.0)
            return 1.0;
        else
            return chebyshevSeriesSum(&integCoeffs_[0], integCoeffs_.size()-1, x);
    }

    double ChebyshevDistro1D::unscaledQuantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::ChebyshevDistro1D::unscaledQuantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return -1.0;
        if (r1 == 1.0)
            return 1.0;
        double q;
        const bool status = findRootUsingBisections(
            UnscaledCdfFunctor1D(*this), r1, -1.0, 1.0, 4.0*DBL_EPSILON, &q);
        assert(status);
        return q;
    }

    double ChebyshevDistro1D::unscaledExceedance(const double x) const
    {
        if (x <= -1.0)
            return 1.0;
        else if (x >= 1.0)
            return 0.0;
        else
            return 1.0 - unscaledCdf(x);
    }

    bool ChebyshevDistro1D::isEqual(const AbsDistribution1D& other) const
    {
        const ChebyshevDistro1D& r = static_cast<const ChebyshevDistro1D&>(other);
        return AbsScalableDistribution1D::isEqual(r) && allCoeffs_ == r.allCoeffs_
            && isPositive_ == r.isPositive_;
    }

    bool ChebyshevDistro1D::write(std::ostream& os) const
    {
        AbsScalableDistribution1D::write(os);
        gs::write_pod_vector(os, allCoeffs_);
        gs::write_pod(os, isPositive_);
        return !os.fail();
    }

    ChebyshevDistro1D* ChebyshevDistro1D::read(const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(gs::ClassId::makeId<ChebyshevDistro1D>());
        current.ensureSameId(id);

        double location, scale;
        if (AbsScalableDistribution1D::read(in, &location, &scale))
        {
            std::vector<double> table;
            gs::read_pod_vector(in, &table);
            bool pos;
            gs::read_pod(in, &pos);
            if (!in.fail() && table.size())
                return new ChebyshevDistro1D(location, scale,
                                             &table[1], table.size()-1, pos);
        }
        distributionReadError(in, classname());
        return 0;
    }
}
