#ifndef NPSTAT_CHEBYSHEVDISTRO1D_HH_
#define NPSTAT_CHEBYSHEVDISTRO1D_HH_

/*!
// \file ChebyshevDistro1D.hh
//
// \brief Statistical distribution constructed using series in
//        Chebyshev polynomials of the first kind
//
// Author: I. Volobouev
//
// February 2023
*/

#include "npstat/nm/ClassicalOrthoPolys1D.hh"

#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    /**
    // A distribution whose density is a series in terms of 
    // Chebyshev polynomials of the first kind. The series
    // coefficients start with degree 1 (note, not 0).
    */
    class ChebyshevDistro1D : public AbsScalableDistribution1D
    {
    public:
        ChebyshevDistro1D(double location, double scale,
                          const double* coeffs, unsigned maxdeg,
                          bool ensurePositivity = true);

        ChebyshevDistro1D(double location, double scale,
                          const std::vector<double>& coeffs,
                          bool ensurePositivity = true);

        inline virtual ChebyshevDistro1D* clone() const
            {return new ChebyshevDistro1D(*this);}

        inline virtual ~ChebyshevDistro1D() {}

        inline bool isPositive() const {return isPositive_;};

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname()
            {return "npstat::ChebyshevDistro1D";}
        static inline unsigned version() {return 1;}
        static ChebyshevDistro1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<ChebyshevDistro1D>;

        inline static int nParameters() {return -1;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;
        double unscaledExceedance(double x) const;

        void checkPositive(bool validatePositivity);
        void setupCoeffs(const double* coeffs, unsigned maxdeg);

        std::vector<double> allCoeffs_;
        std::vector<double> integCoeffs_;
        bool isPositive_;
    };
}

#endif // NPSTAT_CHEBYSHEVDISTRO1D_HH_
