#include "npstat/nm/RecurrenceCoeffsTrGauss.hh"

#include "npstat/stat/johnsonTransform.hh"
#include "npstat/stat/JohnsonOrthoPoly1D.hh"

namespace npstat {
    JohnsonOrthoPoly1D::JohnsonOrthoPoly1D(const AbsDistribution1D& distro,
                                           const unsigned maxDegree,
                                           const unsigned nIntegPoints,
                                           const OrthoPolyMethod m)
        : distro_(distro.clone()), maxdeg_(maxDegree)
    {
        if (maxdeg_)
            initialize(nIntegPoints, m);
        else
        {
            // Just believe that the distro is correctly
            // normalized. We should not be asked for the
            // recurrence coefficients in the future.
            rCoeffs_.reserve(1);
            rCoeffs_.push_back(Private::Recurrence(0.0L, 1.0L));
        }
    }

    void JohnsonOrthoPoly1D::initialize(const unsigned nInteg,
                                        const OrthoPolyMethod m)
    {
        std::unique_ptr<AbsDistributionTransform1D> transform = johnsonTransform(*distro_);
        BackDistroTransform1DFunctor fcn(*transform);
        Private::calcRecurrenceCoeffsTrGauss(fcn, maxdeg_, m, nInteg, &rCoeffs_);
    }

    std::pair<long double,long double>
    JohnsonOrthoPoly1D::recurrenceCoeffs(const unsigned deg) const
    {
        const Private::Recurrence& r = rCoeffs_.at(deg);
        return std::pair<long double,long double>(r.alpha, r.sqrbeta);
    }

    std::pair<long double,long double>
    JohnsonOrthoPoly1D::monicRecurrenceCoeffs(const unsigned deg) const
    {
        const Private::Recurrence& r = rCoeffs_.at(deg);
        return std::pair<long double,long double>(r.alpha, r.beta);
    }

    JohnsonOrthoPoly1D::JohnsonOrthoPoly1D(const JohnsonOrthoPoly1D& r)
        : distro_(r.distro_->clone()),
          rCoeffs_(r.rCoeffs_),
          maxdeg_(r.maxdeg_)
    {
    }

    JohnsonOrthoPoly1D& JohnsonOrthoPoly1D::operator=(const JohnsonOrthoPoly1D& r)
    {
        if (&r != this)
        {
            distro_ = CPP11_auto_ptr<AbsDistribution1D>(r.distro_->clone());
            rCoeffs_ = r.rCoeffs_;
            maxdeg_ = r.maxdeg_;
        }
        return *this;
    }
}
