#ifndef NPSTAT_JOHNSONORTHOPOLY1D_HH_
#define NPSTAT_JOHNSONORTHOPOLY1D_HH_

/*!
// \file JohnsonOrthoPoly1D.hh
//
// \brief Continuous orthonormal polynomial series in one dimension
//        generated for a Johnson curve used as the weight
//
// Author: I. Volobouev
//
// November 2022
*/

#include <vector>

#include "geners/CPP11_auto_ptr.hh"

#include "npstat/nm/AbsClassicalOrthoPoly1D.hh"
#include "npstat/nm/Recurrence.hh"
#include "npstat/nm/OrthoPolyMethod.hh"

#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    class JohnsonOrthoPoly1D : public AbsClassicalOrthoPoly1D
    {
    public:
        /**
        // "nIntegPoints" parameter is the number of integration points
        // for the Gauss-Hermite quadrature.
        */
        JohnsonOrthoPoly1D(const AbsDistribution1D& distro,
                           unsigned maxDegree,
                           unsigned nIntegPoints = 256U,
                           OrthoPolyMethod m = OPOLY_STIELTJES);

        JohnsonOrthoPoly1D(const JohnsonOrthoPoly1D&);
        JohnsonOrthoPoly1D& operator=(const JohnsonOrthoPoly1D&);

        inline virtual ~JohnsonOrthoPoly1D() {}

        inline virtual JohnsonOrthoPoly1D* clone() const
            {return new JohnsonOrthoPoly1D(*this);}

        inline virtual long double weight(const long double x) const
            {return distro_->density(x);}
        inline virtual double xmin() const {return distro_->quantile(0.0);}
        inline virtual double xmax() const {return distro_->quantile(1.0);}
        inline virtual unsigned maxDegree() const {return maxdeg_;}

        virtual std::pair<long double,long double>
        monicRecurrenceCoeffs(unsigned deg) const;

        virtual std::pair<long double,long double>
        recurrenceCoeffs(unsigned deg) const;

    private:
        void initialize(unsigned nIntegPoints, OrthoPolyMethod m);

        CPP11_auto_ptr<AbsDistribution1D> distro_;
        std::vector<Private::Recurrence> rCoeffs_;
        unsigned maxdeg_;
    };
}

#endif // NPSTAT_JOHNSONORTHOPOLY1D_HH_
