#include <cmath>

#include "geners/binaryIO.hh"
#include "npstat/stat/LognormalTransform1D.hh"

namespace npstat {
    LognormalTransform1D::LognormalTransform1D(const double delta,
                                               const double gamma,
                                               const double xi,
                                               const bool positiveSkewness)
        : AbsDistributionTransform1D(4U)
    {
        params_[0] = delta;
        params_[1] = gamma;
        params_[2] = xi;
        params_[3] = positiveSkewness ? 1.0 : 0.0;

        if (delta <= 0.0) throw std::invalid_argument(
            "In npstat::LognormalTransform1D constructor: invalid parameters");
    }

    bool LognormalTransform1D::isIncreasing() const
    {
        return params_[3];
    }

    double LognormalTransform1D::transformForward(const double x, double* dydx) const
    {
        const double delta = params_[0];
        const double gamma = params_[1];
        const double xi = params_[2];
        const bool positiveSkewness = params_[3];

        const double diff = positiveSkewness ? x - xi : -x - xi;
        if (diff <= 0.0) throw std::invalid_argument(
            "In npstat::LognormalTransform1D::transformForward: "
            "argument is out of range");
        if (dydx)
        {
            *dydx = delta/diff;
            if (!positiveSkewness)
                *dydx *= -1.0;
        }
        return gamma + delta*log(diff);
    }

    double LognormalTransform1D::transformBack(const double z) const
    {
        const double tmp = exp((z - params_[1])/params_[0]) + params_[2];
        const bool positiveSkewness = params_[3];
        return positiveSkewness ? tmp : -tmp;
    }

    bool LognormalTransform1D::isEqual(const AbsDistributionTransform1D& o) const
    {
        const LognormalTransform1D& r = static_cast<const LognormalTransform1D&>(o);
        for (unsigned i=0; i<4U; ++i)
            if (!(params_[i] == r.params_[i]))
                return false;
        return true;
    }

    bool LognormalTransform1D::write(std::ostream& os) const
    {
        gs::write_pod_array(os, params_, 4UL);
        return !os.fail();
    }

    LognormalTransform1D* LognormalTransform1D::read(const gs::ClassId& id,
                                                     std::istream& is)
    {
        static const gs::ClassId current(gs::ClassId::makeId<LognormalTransform1D>());
        current.ensureSameId(id);

        double tmp[4];
        gs::read_pod_array(is, tmp, 4UL);
        if (is.fail()) throw gs::IOReadFailure(
            "In npstat::LognormalTransform1D::read: input stream failure");
        return new LognormalTransform1D(tmp[0], tmp[1], tmp[2], tmp[3]);
    }
}
