#include <cmath>
#include <algorithm>

#include "npstat/nm/ClassicalOrthoPolys1D.hh"
#include "npstat/nm/ArrayND.hh"

#include "npstat/stat/NeymanOSDE1D.hh"

static double sprod_dld(const double* p1, const long double* p2, const unsigned sz)
{
    long double sum = 0.0L;
    for (unsigned i=0; i<sz; ++i)
        sum += *p1++ * *p2++;
    return sum;
}

namespace npstat {
    void NeymanOSDE1D::moveCoordsToStandardInterval()
    {
        scale_ = (xmax_ - xmin_)/2.0;
        shift_ = (xmax_ + xmin_)/2.0;
        for (unsigned long i=0; i<nCoords_; ++i)
        {
            const double x = coords_[i];
            assert(x >= xmin_);
            assert(x <= xmax_);

            double xt = (x - shift_)/scale_;
            if (xt < -1.0)
                xt = -1.0;
            else if (xt > 1.0)
                xt = 1.0;
            coords_[i] = xt;
        }
    }

    void NeymanOSDE1D::fillIntegralTable(const unsigned maxdeg) const
    {
        if (maxdeg > currentMaxDeg_)
        {
            LegendreOrthoPoly1D lpoly;

            std::vector<double> coeffs(maxdeg+1U, 0.0);
            std::vector<double> buf(maxdeg+2U);
            double* integCoeffs = &buf[0];

            for (unsigned ideg=currentMaxDeg_+1U; ideg<=maxdeg; ++ideg)
            {
                coeffs[ideg-1U] = 0.0;
                coeffs[ideg] = 1.0;
                lpoly.integralCoeffs(&coeffs[0], ideg, integCoeffs);
                const unsigned degp1 = ideg + 1U;
                for (unsigned long ipt=0; ipt<nCoords_; ++ipt)
                    integs_.push_back(lpoly.series(integCoeffs, degp1, coords_[ipt]));
            }
            currentMaxDeg_ = maxdeg;
            assert(integs_.size() == currentMaxDeg_*nCoords_);
        }
    }

    double NeymanOSDE1D::cdf(const double* coeffs, const unsigned ncoeffs,
                             const unsigned long ipt) const
    {
        assert(ipt < nCoords_);
        if (ncoeffs > currentMaxDeg_)
            fillIntegralTable(ncoeffs);
        long double sum = (coords_[ipt] + 1.0)/2.0;
        if (ncoeffs)
        {
            assert(coeffs);
            for (unsigned i=0; i<ncoeffs; ++i)
                sum += coeffs[i]*integs_[i*nCoords_ + ipt];
        }
        return sum;
    }

    void NeymanOSDE1D::fillDerivativeTables(const unsigned maxdeg) const
    {
        const unsigned currentDeg = firstDeriv_.size();
        if (maxdeg > currentDeg)
        {
            ShiftedLegendreOrthoPoly1D poly;

            std::vector<double> coeffs(maxdeg+1U, 0.0);

            for (unsigned ideg=currentDeg+1U; ideg<=maxdeg; ++ideg)
            {
                std::vector<double> derivCoeffs(ideg);
                coeffs[ideg-1U] = 0.0;
                coeffs[ideg] = 1.0;
                poly.derivativeCoeffs(&coeffs[0], ideg, &derivCoeffs[0]);
                firstDeriv_.push_back(derivCoeffs);

                std::vector<double> secondDerivC(std::max(ideg-1U, 1U));
                poly.derivativeCoeffs(&derivCoeffs[0], ideg-1U, &secondDerivC[0]);
                secondDeriv_.push_back(secondDerivC);
            }

            assert(firstDeriv_.size() == maxdeg);
            assert(secondDeriv_.size() == maxdeg);
        }
    }

    double NeymanOSDE1D::chiSquare(
        const double* coeffs, const unsigned nCoeffs,
        const double* shrinkages, const unsigned nShrinkages,
        double* polyStatistics,
        double* gradient,
        Matrix<double>* hessian) const
    {
        assert(nCoeffs);
        assert(nShrinkages >= nCoeffs);
        assert(coeffs);
        assert(shrinkages);
        assert(gradient);

        // Make sure that all shrinkages are non-negative
        unsigned shrCount = 0;
        for (unsigned ishr=0U; ishr<nShrinkages; ++ishr)
        {
            assert(shrinkages[ishr] >= 0.0);
            if (shrinkages[ishr] > 0.0)
                ++shrCount;
        }
        assert(shrCount >= nCoeffs);

        fillIntegralTable(nCoeffs);
        fillDerivativeTables(nShrinkages);
        
        ShiftedLegendreOrthoPoly1D poly;

        std::vector<long double> polySums(nShrinkages+1U, 0.0L);
        std::vector<long double> polyBuf(nShrinkages+1U);
        Matrix<long double> derMat(nShrinkages, nCoeffs, 0);
        ArrayND<long double> sderTensor(nShrinkages, nCoeffs, nCoeffs);
        if (hessian)
            sderTensor.clear();
        double dp2dy2 = 0.0;

        for (unsigned long ipt=0; ipt<nCoords_; ++ipt)
        {
            const double y = cdf(coeffs, nCoeffs, ipt);
            poly.allpoly(y, &polyBuf[0], nShrinkages);

            for (unsigned ideg=1U; ideg<=nShrinkages; ++ideg)
            {
                polySums[ideg] += polyBuf[ideg];
                const unsigned idegm1 = ideg-1U;
                const double dpdy = sprod_dld(&firstDeriv_[idegm1][0],
                                              &polyBuf[0], firstDeriv_[idegm1].size());
                if (hessian)
                    dp2dy2 = sprod_dld(&secondDeriv_[idegm1][0],
                                       &polyBuf[0], secondDeriv_[idegm1].size());
                for (unsigned icoeff=0U; icoeff<nCoeffs; ++icoeff)
                {
                    derMat[idegm1][icoeff] += dpdy*integs_[icoeff*nCoords_ + ipt];
                    if (hessian)
                        for (unsigned m=0U; m<nCoeffs; ++m)
                            sderTensor(idegm1, icoeff, m) += dp2dy2*integs_[icoeff*nCoords_ + ipt]*integs_[m*nCoords_ + ipt];
                }
            }
        }

        // Calculate polynomial statistics
        const double sqrtn = std::sqrt(static_cast<double>(nCoords_));
        for (unsigned ideg=1U; ideg<=nShrinkages; ++ideg)
            polyStatistics[ideg-1U] = polySums[ideg]/sqrtn;

        // Calculate the gradient
        for (unsigned icoeff=0U; icoeff<nCoeffs; ++icoeff)
        {
            long double sum = 0.0L;
            for (unsigned ideg=0; ideg<nShrinkages; ++ideg)
                sum += polyStatistics[ideg]*shrinkages[ideg]*derMat[ideg][icoeff];
            gradient[icoeff] = 2.0*sum/sqrtn;
        }

        // Calculate the Hessian
        if (hessian)
        {
            hessian->resize(nCoeffs, nCoeffs);
            for (unsigned j=0; j<nCoeffs; ++j)
                for (unsigned k=0; k<=j; ++k)
                {
                    long double sum = 0.0L;
                    for (unsigned ideg=0; ideg<nShrinkages; ++ideg)
                        sum += shrinkages[ideg]*(derMat[ideg][j]*derMat[ideg][k] +
                                                 polySums[ideg+1U]*sderTensor(ideg, j, k));
                    (*hessian)[j][k] = 2.0*sum/nCoords_;
                    if (j != k)
                        (*hessian)[k][j] = (*hessian)[j][k];
                }
        }

        // Finally, calculate and return the chi-square
        long double chisq = 0.0L;
        for (unsigned ishr=0U; ishr<nShrinkages; ++ishr)
            chisq += shrinkages[ishr]*polyStatistics[ishr]*polyStatistics[ishr];
        return chisq;
    }

    Matrix<double> NeymanOSDE1D::sensitivityMatrix(
        const double* coeffs, const unsigned nCoeffs,
        const unsigned nShrinkages) const
    {
        assert(nCoeffs);
        assert(nShrinkages >= nCoeffs);
        assert(coeffs);

        fillIntegralTable(nCoeffs);
        fillDerivativeTables(nShrinkages);
        
        ShiftedLegendreOrthoPoly1D poly;

        std::vector<long double> polyBuf(nShrinkages+1U);
        Matrix<long double> derMat(nShrinkages, nCoeffs, 0);

        for (unsigned long ipt=0; ipt<nCoords_; ++ipt)
        {
            const double y = cdf(coeffs, nCoeffs, ipt);
            poly.allpoly(y, &polyBuf[0], nShrinkages);

            for (unsigned ideg=1U; ideg<=nShrinkages; ++ideg)
            {
                const unsigned idegm1 = ideg-1U;
                const double dpdy = sprod_dld(&firstDeriv_[idegm1][0],
                                              &polyBuf[0], firstDeriv_[idegm1].size());
                for (unsigned icoeff=0U; icoeff<nCoeffs; ++icoeff)
                    derMat[idegm1][icoeff] += dpdy*integs_[icoeff*nCoords_ + ipt];
            }
        }

        const double sqrtn = std::sqrt(static_cast<double>(nCoords_));
        return derMat/sqrtn;
    }

    void NeymanOSDE1D::sampleCoeffs(double* coeffs, const unsigned maxdeg) const
    {
        if (maxdeg)
        {
            assert(coeffs);
            LegendreOrthoPoly1D lpoly;
            std::vector<double> tmp(maxdeg+1U);
            lpoly.sampleCoeffs(&coords_[0], nCoords_, &tmp[0], maxdeg);
            const double* from = &tmp[1];
            for (unsigned i=0; i<maxdeg; ++i)
                *coeffs++ = *from++;
        }
    }

    LegendreDistro1D NeymanOSDE1D::densityEstimate(
        const double* coeffs, const unsigned nCoeffs) const
    {
        return LegendreDistro1D(shift_, scale_, coeffs, nCoeffs, false);
    }
}
