#ifndef NPSTAT_NEYMANOSDE1DRESULT_HH_
#define NPSTAT_NEYMANOSDE1DRESULT_HH_

/*!
// \file NeymanOSDE1DResult.hh
//
// \brief Class for representing results of the Neyman OSDE
//        optimization algorithms
//
// Author: I. Volobouev
//
// March 2023
*/

#include <vector>

namespace npstat {
    enum NeymanOSDE1DMinimizerCode {
        NOSDE_NOT_OPTIMIZED = 0,
        NOSDE_NEWTON,
        NOSDE_GRADIENT_DESCENT,
        NOSDE_RANDOM_SEARCH,
        NOSDE_MINUIT,
        NOSDE_MINUIT_IMPROVED,
        NOSDE_OTHER,
        NOSDE_N_MINIMIZER_CODES
    };

    const char* neymanOSDEMinimizerName(unsigned minimizerCode);
    
    class NeymanOSDE1DResult
    {
    public:
        inline NeymanOSDE1DResult(const std::vector<double>& i_coeffs,
                                  const std::vector<double>& polyStats,
                                  const double i_chiSquare,
                                  const double i_maxAbsGradient,
                                  const unsigned i_minimizerCode,
                                  const unsigned i_nIterations,
                                  const bool i_converged,
                                  const bool i_badHessian)
            : coeffs_(i_coeffs),
              polyStatistics_(polyStats),
              chiSquare_(i_chiSquare),
              maxAbsGradient_(i_maxAbsGradient),
              minimizerCode_(i_minimizerCode),
              nIterations_(i_nIterations),
              converged_(i_converged),
              badHessian_(i_badHessian)
        {
        }                          

        inline void setMinimizerCode(const unsigned c) {minimizerCode_ = c;}

        inline const std::vector<double>& coeffs() const {return coeffs_;}
        inline const std::vector<double>& polyStatistics() const
            {return polyStatistics_;}
        inline double chiSquare() const {return chiSquare_;}
        inline double maxAbsGradient() const {return maxAbsGradient_;}
        inline unsigned minimizerCode() const {return minimizerCode_;}
        inline unsigned nIterations() const {return nIterations_;}
        inline bool converged() const {return converged_;}
        inline bool badHessian() const {return badHessian_;}

        inline bool operator<(const NeymanOSDE1DResult& r) const
        {
            if (chiSquare_ < r.chiSquare_) return true;
            if (r.chiSquare_ < chiSquare_) return false;
            if (maxAbsGradient_ < r.maxAbsGradient_) return true;
            if (r.maxAbsGradient_ < maxAbsGradient_) return false;
            if (converged_ && !r.converged_) return true;
            if (!converged_ && r.converged_) return false;
            return false;
        }

    private:
        std::vector<double> coeffs_;
        std::vector<double> polyStatistics_;
        double chiSquare_;
        double maxAbsGradient_;
        unsigned minimizerCode_;
        unsigned nIterations_;
        bool converged_;
        bool badHessian_;
    };
}

#endif // NPSTAT_NEYMANOSDE1DRESULT_HH_
