#include "npstat/nm/ClassicalOrthoPolys1D.hh"

#include "npstat/stat/PLogliOSDE1D.hh"

namespace npstat {
    void PLogliOSDE1D::moveCoordsToStandardInterval()
    {
        scale_ = (xmax_ - xmin_)/2.0;
        shift_ = (xmax_ + xmin_)/2.0;
        for (unsigned long i=0; i<nCoords_; ++i)
        {
            const double x = coords_[i];
            assert(x >= xmin_);
            assert(x <= xmax_);

            double xt = (x - shift_)/scale_;
            if (xt < -1.0)
                xt = -1.0;
            else if (xt > 1.0)
                xt = 1.0;
            coords_[i] = xt;
        }
    }

    void PLogliOSDE1D::fillDerivativeTables(const unsigned maxdeg) const
    {
        if (maxdeg > currentMaxDeg_)
        {
            LegendreOrthoPoly1D lpoly;

            for (unsigned ideg=currentMaxDeg_+1U; ideg<=maxdeg; ++ideg)
                for (unsigned long ipt=0; ipt<nCoords_; ++ipt)
                    derivs_.push_back(lpoly.poly(ideg, coords_[ipt]));
            currentMaxDeg_ = maxdeg;
            assert(derivs_.size() == currentMaxDeg_*nCoords_);
        }
    }

    double PLogliOSDE1D::pseudoLogli(
        const double* i_coeffs, const unsigned nCoeffs, double* gradient) const
    {
        assert(nCoeffs);
        assert(i_coeffs);
        assert(gradient);

        fillDerivativeTables(nCoeffs);

        LegendreOrthoPoly1D poly;
        long double logli = 0.0L;
        std::vector<long double> grad(nCoeffs, 0.0L);
        std::vector<double> coeffs(nCoeffs + 1U);
        coeffs[0] = 0.5;
        for (unsigned i=0; i<nCoeffs; ++i)
            coeffs[i+1U] = i_coeffs[i];

        for (unsigned long ipt=0; ipt<nCoords_; ++ipt)
        {
            const double density = poly.series(&coeffs[0], nCoeffs, coords_[ipt]);
            logli += trlog_(density);

            const double deriv = trlog_.derivative(density);
            for (unsigned ic=0; ic<nCoeffs; ++ic)
                grad[ic] += deriv*derivs_[ic*nCoords_ + ipt];
        }

        for (unsigned i=0; i<nCoeffs; ++i)
            gradient[i] = grad[i];
        return logli;
    }

    void PLogliOSDE1D::sampleCoeffs(double* coeffs, const unsigned maxdeg) const
    {
        if (maxdeg)
        {
            assert(coeffs);
            LegendreOrthoPoly1D lpoly;
            std::vector<double> tmp(maxdeg+1U);
            lpoly.sampleCoeffs(&coords_[0], nCoords_, &tmp[0], maxdeg);
            const double* from = &tmp[1];
            for (unsigned i=0; i<maxdeg; ++i)
                *coeffs++ = *from++;
        }
    }

    LegendreDistro1D PLogliOSDE1D::densityEstimate(
        const double* coeffs, const unsigned nCoeffs) const
    {
        return LegendreDistro1D(shift_, scale_, coeffs, nCoeffs, false);
    }
}
