#ifndef NPSTAT_PLOGLIOSDE1D_HH_
#define NPSTAT_PLOGLIOSDE1D_HH_

/*!
// \file PLogliOSDE1D.hh
//
// \brief OSDE based on pseudo-log-likelihood
//
// Author: I. Volobouev
//
// March 2023
*/

#include <vector>
#include <limits>

#include "npstat/nm/Matrix.hh"
#include "npstat/nm/TruncatedLog.hh"

#include "npstat/stat/LegendreDistro1D.hh"

namespace npstat {
    class PLogliOSDE1D
    {
    public:
        template<typename Numeric>
        PLogliOSDE1D(const Numeric* coords, unsigned long lenCoords,
                     double xmin, double xmax, double logTruncationX,
                     unsigned nLogTaylorTerms);

        inline double xmin() const {return xmin_;}
        inline double xmax() const {return xmax_;}
        inline unsigned long sampleSize() const {return nCoords_;}
        inline double coord(unsigned long i) const
            {return scale_*coords_.at(i) + shift_;}
        inline double logTruncationX() const {return trlog_.truncationX();}
        inline unsigned logTaylorTerms() const {return trlog_.nTaylorTerms();}

        // coeffs   -- array of coefficients, starting with degree 1.
        // nCoeffs  -- length of the array "coeffs".
        //
        // The function returns the negative pseudo-log-likelihood value.
        // Also, the following quantities are filled out on exit:
        //
        // gradient --  pseudo-log-likelihood gradient w.r.t. coeffs. The
        //              length of this array should be at least "nCoeffs".
        //
        double pseudoLogli(const double* coeffs, unsigned nCoeffs,
                           double* gradient) const;

        // Get the expectation values of the Legendre polynomials
        // for the sample, starting with degree 1. The length of
        // the array "coeffs" should be maxdeg.
        void sampleCoeffs(double* coeffs, unsigned maxdeg) const;

        // Once the coefficients have been determined, get the
        // density estimate for the original sample. Note that
        // this density estimate will not be guaranteed positive.
        // The coefficients start with degree 1.
        LegendreDistro1D densityEstimate(const double* coeffs,
                                         unsigned nCoeffs) const;

        // Check whether the given set of coefficients creates
        // a series which can be used as a bona fide density.
        // The coefficients start with degree 1.
        inline bool isPositive(const double* coeffs,
                               const unsigned nCoeffs) const
            {return densityEstimate(coeffs, nCoeffs).isPositive();}

    private:
        void moveCoordsToStandardInterval();
        void fillDerivativeTables(unsigned maxdeg) const;

        // Coordinates will be shifted and scaled to the [-1, 1] interval
        std::vector<double> coords_;
        unsigned long nCoords_;
        double xmin_;
        double xmax_;
        double scale_;
        double shift_;
        TruncatedLog<double> trlog_;

        mutable std::vector<double> derivs_;
        mutable unsigned currentMaxDeg_;
    };
}

#include "npstat/stat/PLogliOSDE1D.icc"

#endif // NPSTAT_PLOGLIOSDE1D_HH_
