#ifndef NPSTAT_TRIG2GOFTEST1D_HH_
#define NPSTAT_TRIG2GOFTEST1D_HH_

/*!
// \file Trig2GOFTest1D.hh
//
// \brief Goodness-of-fit test based on the trigonometric expansion
//        of the empirical comparison density
//
// Author: I. Volobouev
//
// May 2023
*/

#include "npstat/stat/AbsSmoothGOFTest1D.hh"

namespace npstat {
    class Trig2GOFTest1D : public AbsSmoothGOFTest1D
    {
    public:
        // The mask, if provided, should contain nTrig2Terms
        // elements. Put some number in the mask in order to
        // discard the corresponding term and put 0 there to
        // keep it. The ordering of terms is as follows:
        //
        // sin(2 Pi x)   -> mask[0]
        // cos(2 Pi x)   -> mask[1]
        // sin(4 Pi x)   -> mask[2]
        // cos(4 Pi x)   -> mask[3]
        // ........................
        // sin(2 k Pi x) -> mask[2 k - 2]
        // cos(2 k Pi x) -> mask[2 k - 1]
        // ........................
        //
        // Maximum value of k equals nTrig2Terms / 2 (integer division)
        // in case nTrig2Terms is even. In case nTrig2Terms is odd,
        // it is nTrig2Terms / 2 + 1, but only the sine term is used.
        //
        inline Trig2GOFTest1D(const AbsDistribution1D& distro,
                              const unsigned nTrig2Terms,
                              const unsigned char* mask=0,
                              const unsigned lenMask=0)
            : AbsSmoothGOFTest1D(distro, nTrig2Terms, mask, lenMask) {}

        inline Trig2GOFTest1D(const AbsDistribution1D& distro,
                              const std::vector<int>& mask)
            : AbsSmoothGOFTest1D(distro, mask) {}

        inline virtual ~Trig2GOFTest1D() {}

        inline virtual std::string shortName() const
            {return testNameWithMask("Trig2");}

        virtual unsigned normalizedDeviations(
            const double* data, unsigned long sz,
            bool isDataSorted,
            double* deviations, unsigned lenDeviations) const;

        virtual unsigned normalizedDeviations(
            const float* data, unsigned long sz,
            bool isDataSorted,
            double* deviations, unsigned lenDeviations) const;

    private:
        template<typename Numeric>
        unsigned normalizedDeviationsHelper(
            const Numeric* data, unsigned long lenData,
            double* deviations) const;
    };
}

#include "npstat/stat/Trig2GOFTest1D.icc"

#endif // NPSTAT_TRIG2GOFTEST1D_HH_
