#include <cmath>
#include <cassert>

#include "npstat/stat/TwoDistros1DFunctors.hh"

#include "npstat/nm/ClassicalOrthoPolys1D.hh"
#include "npstat/nm/GaussLegendreQuadrature.hh"
#include "npstat/nm/RectangleQuadrature1D.hh"

namespace {
    class ComparisonDensityProdFcn : public npstat::Functor1<double, double>
    {
    public:
        inline ComparisonDensityProdFcn(
            const npstat::ComparisonDensityFunctor1D& cd, const unsigned deg)
            : cd_(cd), deg_(deg) {}

        inline virtual ~ComparisonDensityProdFcn() {}

        inline virtual double operator()(const double& x) const
            {return cd_(x)*poly_.poly(deg_, x);}

    private:
        const npstat::ComparisonDensityFunctor1D& cd_;
        npstat::ShiftedLegendreOrthoPoly1D poly_;
        unsigned deg_;
    };

    class ComparisonDensityTrigFcn : public npstat::Functor1<double, double>
    {
    public:
        inline ComparisonDensityTrigFcn(
            const npstat::ComparisonDensityFunctor1D& cd,
            const unsigned k, const bool isCos)
            : cd_(cd), k_(k), isCos_(isCos) {}

        inline virtual ~ComparisonDensityTrigFcn() {}

        inline virtual double operator()(const double& x) const
        {
            const double y = k_*M_PI*x;
            double trig;
            if (isCos_)
                trig = std::cos(y);
            else
                trig = std::sin(y);
            return cd_(x)*trig;
        }

    private:
        const npstat::ComparisonDensityFunctor1D& cd_;
        unsigned k_;
        bool isCos_;
    };
}

namespace npstat {
    double ComparisonDensityFunctor1D::legendreCoeff(
        const unsigned degree) const
    {
        GaussLegendreQuadrature glq(1024U);
        ComparisonDensityProdFcn fcn(*this, degree);
        return glq.integrate(fcn, 0.0, 1.0);
    }

    double ComparisonDensityFunctor1D::trigCoeff(
        const unsigned k, const bool isCosine, const unsigned nInteg) const
    {
        static const double sqr2 = 1.414213562373095049;

        RectangleQuadrature1D rq(nInteg);
        ComparisonDensityTrigFcn fcn(*this, 2U*k, isCosine);
        return sqr2*rq.integrate(fcn, 0.0, 1.0);
    }

    double ComparisonDensityFunctor1D::cosCoeff(
        const unsigned k, const unsigned nInteg) const
    {
        static const double sqr2 = 1.414213562373095049;

        RectangleQuadrature1D rq(nInteg);
        ComparisonDensityTrigFcn fcn(*this, k, true);
        return sqr2*rq.integrate(fcn, 0.0, 1.0);
    }

    void ComparisonDensityFunctor1D::legendreSeriesCoeffs(
        double* coeffs, const unsigned maxDeg) const
    {
        assert(coeffs);

        ShiftedLegendreOrthoPoly1D poly;
        GaussLegendreQuadrature glq(1024U);
        poly.extWeightAverages(*this, glq, coeffs, maxDeg);
    }
}
