#ifndef NPSTAT_TWODISTROS1DFUNCTORS_HH_
#define NPSTAT_TWODISTROS1DFUNCTORS_HH_

/*!
// \file TwoDistros1DFunctors.hh
//
// \brief A few functors used to study GoF tests
//
// Author: I. Volobouev
//
// October 2022
*/

#include <stdexcept>

#include "npstat/stat/AbsTwoDistros1DFunctor.hh"

namespace npstat {
    class CdfDeltaFunctor1D : public AbsTwoDistros1DFunctor
    {
    public:
        inline CdfDeltaFunctor1D(const AbsDistribution1D& nullHypothesis,
                                 const AbsDistribution1D& alternative)
            : AbsTwoDistros1DFunctor(nullHypothesis, alternative) {}

        inline virtual ~CdfDeltaFunctor1D() {}

        inline virtual double operator()(const double& x) const
        {
            if (x < 0.0 || x > 1.0)
                return 0.0;
            else
                return altPtr_->cdf(nullPtr_->quantile(x)) - x;
        }
    };

    class ComparisonDensityFunctor1D : public AbsTwoDistros1DFunctor
    {
    public:
        inline ComparisonDensityFunctor1D(
            const AbsDistribution1D& nullHypothesis,
            const AbsDistribution1D& alternative)
            : AbsTwoDistros1DFunctor(nullHypothesis, alternative) {}

        inline virtual ~ComparisonDensityFunctor1D() {}

        inline virtual double operator()(const double& x) const
        {
            if (x < 0.0 || x > 1.0)
                return 0.0;
            else
            {
                const double y = nullPtr_->quantile(x);
                const double nulldens = nullPtr_->density(y);
                if (nulldens <= 0.0) throw std::runtime_error(
                    "In npstat::ComparisonDensityFunctor1D::operator(): "
                    "baseline density is not positive");
                return altPtr_->density(y)/nulldens;
            }
        }

        /**
        // Coefficient of expansion of this function into
        // shifted Legendre polynomials (that is, Legendre
        // polynomials orthonormal on [0, 1]).
        */
        double legendreCoeff(unsigned degree) const;

        /**
        // All coefficients of the Legendre polynomial expansion
        // of the density, up to the degree requested. The array
        // "coeffs" should have at least maxDeg+1 elements.
        */
        void legendreSeriesCoeffs(double* coeffs, unsigned maxDeg) const;

        /**
        // Coefficient of expansion of this function into
        // trigonometric functions, that is, sqrt(2)*sin(2 k Pi x)
        // and sqrt(2)*cos(2 k Pi x). The integration is performed
        // using the rectangle integration rule, with nInteg
        // intervals. nInteg should be substantially larger than k,
        // something like 32*k should work reasonably well.
        */
        double trigCoeff(unsigned k, bool isCosine, unsigned nInteg) const;

        /**
        // Coefficient of expansion of this function into cosine
        // basis, that is, sqrt(2)*cos(k Pi x). The integration
        // is performed using the rectangle integration rule, with
        // nInteg intervals. nInteg should be substantially larger
        // than k, something like 16*k should work reasonably well.
        */
        double cosCoeff(unsigned k, unsigned nInteg) const;

#ifdef SWIG
    public:
        inline std::vector<double>
        legendreSeriesCoeffs2(const unsigned maxDeg) const
        {
            std::vector<double> tmp(maxDeg+1U);
            legendreSeriesCoeffs(&tmp[0], maxDeg);
            return tmp;
        }
#endif // SWIG
    };
}

#endif // NPSTAT_TWODISTROS1DFUNCTORS_HH_
