#include <cassert>
#include <stdexcept>

#include "npstat/stat/JohnsonCurves.hh"
#include "npstat/stat/Distributions1D.hh"

#include "npstat/stat/AsinhTransform1D.hh"
#include "npstat/stat/LogRatioTransform1D.hh"
#include "npstat/stat/LocScaleTransform1D.hh"
#include "npstat/stat/LognormalTransform1D.hh"
#include "npstat/stat/TransformedDistribution1D.hh"

#include "npstat/stat/johnsonTransform.hh"

template<class Distro>
static inline void getJohnsonEffLambdaAndXi(
    const Distro& distro, double* plam, double* pxi)
{
    assert(plam);
    assert(pxi);
    const double mu = distro.location();
    const double sigma = distro.scale();
    const double lam = distro.getLambda();
    const double xi = distro.getXi();
    *plam = lam*sigma;
    *pxi = xi*sigma + mu;
}

namespace npstat {
    std::unique_ptr<AbsDistributionTransform1D> johnsonTransform(
        const AbsDistribution1D& ref)
    {
        const TransformedDistribution1D* trd =
            dynamic_cast<const TransformedDistribution1D*>(&ref);
        if (trd)
        {
            const Gauss1D* gauss = dynamic_cast<const Gauss1D*>(
                &trd->getUnderlyingDistro());
            if (gauss)
                if (gauss->location() == 0.0 && gauss->scale() == 1.0)
                    return std::unique_ptr<AbsDistributionTransform1D>(
                        trd->getTransform().clone());
        }

        const JohnsonSu* jsu = dynamic_cast<const JohnsonSu*>(&ref);
        if (jsu && jsu->isValid())
        {
            double lam, xi;
            getJohnsonEffLambdaAndXi(*jsu, &lam, &xi);
            return std::unique_ptr<AbsDistributionTransform1D>(
                new AsinhTransform1D(jsu->getDelta(), lam,
                                     jsu->getGamma(), xi));
        }

        const JohnsonSb* jsb = dynamic_cast<const JohnsonSb*>(&ref);
        if (jsb && jsb->isValid())
        {
            double lam, xi;
            getJohnsonEffLambdaAndXi(*jsb, &lam, &xi);
            return std::unique_ptr<AbsDistributionTransform1D>(
                new LogRatioTransform1D(jsb->getDelta(), lam,
                                        jsb->getGamma(), xi));
        }

        const Gauss1D* gauss = dynamic_cast<const Gauss1D*>(&ref);
        if (gauss)
            return std::unique_ptr<AbsDistributionTransform1D>(
                new LocScaleTransform1D(gauss->location(), gauss->scale()));

        const LogNormal* lgn = dynamic_cast<const LogNormal*>(&ref);
        if (lgn)
        {
            if (lgn->isGaussian())
                return std::unique_ptr<AbsDistributionTransform1D>(
                    new LocScaleTransform1D(lgn->location(), lgn->scale()));
            else
                return std::unique_ptr<AbsDistributionTransform1D>(
                    new LognormalTransform1D(lgn->getDelta(), lgn->getGamma(),
                                             lgn->getXi(), lgn->skewness() > 0.0));
        }

        const JohnsonSystem* jsys = dynamic_cast<const JohnsonSystem*>(&ref);
        if (jsys)
        {
            const double mu = jsys->location();
            const double sigma = jsys->scale();

            switch (jsys->curveType())
            {
            case JohnsonSystem::GAUSSIAN:
            {
                const Gauss1D distro(mu, sigma);
                return johnsonTransform(distro);
            }
            break;

            case JohnsonSystem::LOGNORMAL:
            {
                const LogNormal distro(mu, sigma, jsys->skewness());
                return johnsonTransform(distro);
            }
            break;

            case JohnsonSystem::SU:
            {
                const JohnsonSu distro(mu, sigma, jsys->skewness(),
                                       jsys->kurtosis());
                return johnsonTransform(distro);
            }
            break;

            case JohnsonSystem::SB:
            {
                const JohnsonSb distro(mu, sigma, jsys->skewness(),
                                       jsys->kurtosis());
                return johnsonTransform(distro);
            }
            break;

            case JohnsonSystem::INVALID:
            {
                throw std::invalid_argument(
                    "In npstat::johnsonTransform: input distribution has "
                    "impossible combination of skewness and kurtosis");
            }
            break;

            default:
            {
                assert(!"Unhandled switch case in npstat::johnsonTransform. "
                       "This is a bug. Please report");
            }
            }
        }

        throw std::invalid_argument(
            "In npstat::johnsonTransform: input distribution is not "
            "a member of Johnson system of frequency curves");
        return std::unique_ptr<AbsDistributionTransform1D>();
    }
}
